@@ -2,7 +2,7 @@ package Net::DNS::RR::SVCB;
 
 use strict;
 use warnings;
-our $VERSION = (qw$Id: SVCB.pm 1847 2021-08-11 10:02:44Z willem $)[2];
+our $VERSION = (qw$Id: SVCB.pm 1965 2024-02-14 09:19:32Z willem $)[2];
 
 use base qw(Net::DNS::RR);
 
@@ -15,8 +15,6 @@ Net::DNS::RR::SVCB - DNS SVCB resource record
 
 use integer;
 
-use Carp;
-use MIME::Base64;
 use Net::DNS::DomainName;
 use Net::DNS::RR::A;
 use Net::DNS::RR::AAAA;
@@ -31,27 +29,31 @@ my %keybyname = (
 	ipv4hint	  => 'key4',
 	ech		  => 'key5',
 	ipv6hint	  => 'key6',
+	dohpath		  => 'key7',				# draft-schwartz-svcb-dns
+	ohttp		  => 'key8',				# draft-pauly-ohai-svcb-config
 	);
 
 
 sub _decode_rdata {			## decode rdata from wire-format octet string
-	my $self = shift;
-	my ( $data, $offset ) = @_;
+	my ( $self, $data, $offset ) = @_;
 
-	my $rdata = substr $$data, $offset, $self->{rdlength};
-	$self->{SvcPriority} = unpack( 'n', $rdata );
-
-	my $index;
-	( $self->{TargetName}, $index ) = Net::DNS::DomainName->decode( \$rdata, 2 );
+	my $limit = $self->{rdlength};
+	my $rdata = substr $$data, $offset, $limit;
+	$self->{SvcPriority} = unpack 'n', $rdata;
+	( $self->{TargetName}, $offset ) = Net::DNS::DomainName->decode( \$rdata, 2 );
 
 	my $params = $self->{SvcParams} = [];
-	my $limit  = length($rdata) - 3;
-	while ( $index < $limit ) {
-		my ( $key, $size ) = unpack( "\@$index n2", $rdata );
-		push @$params, ( $key, substr $rdata, $index + 4, $size );
-		$index += ( $size + 4 );
+	while ( ( my $start = $offset + 4 ) <= $limit ) {
+		my ( $key, $size ) = unpack( "\@$offset n2", $rdata );
+		my $next = $start + $size;
+		last if $next > $limit;
+		push @$params, ( $key, substr $rdata, $start, $size );
+		$offset = $next;
+	}
+	unless ( $offset == $limit ) {
+		$self->{unparsed} = substr $rdata, $offset;
+		die $self->type . ': corrupt RDATA';
 	}
-	die $self->type . ': corrupt RDATA' unless $index == length($rdata);
 	return;
 }
 
@@ -75,45 +77,52 @@ sub _format_rdata {			## format rdata portion of RR string.
 	my $self = shift;
 
 	my $priority = $self->{SvcPriority};
+	my $target   = $self->{TargetName}->string;
 	my $params   = $self->{SvcParams} || [];
-	return ( $priority, $self->{TargetName}->string ) unless scalar @$params;
+	return ( $priority, $target ) unless $priority;
 
+	my $encode = $self->{TargetName}->encode();
+	my $length = 2 + length $encode;
+	my @target = grep {length} split /(\S{32})/, unpack 'H*', $encode;
 	my @rdata  = unpack 'H4', pack 'n', $priority;
-	my $target = $self->{TargetName}->encode();
-	my $length = 2 + length $target;
-	my @target = split /(\S{32})/, unpack 'H*', $target;
-	push @rdata, $length > 18 ? "\t; $priority\n" : (), @target;
-	push @rdata, join '', "\t; ", $self->{TargetName}->string, "\n" if $length > 3;
+	push @rdata, "\t; $priority\n";
+	push @rdata, shift @target;
+	push @rdata, join '', "\t; ", substr( $target, 0, 40 ), "\n";
+	push @rdata, @target;
 
 	my @params = @$params;
 	while (@params) {
 		my $key = shift @params;
 		my $val = shift @params;
 		push @rdata, "\n";
-		push @rdata, "; key$key=...\n" if $key > 15;
-		push @rdata, unpack 'H4H4', pack( 'n2', $key, length $val );
+		push @rdata, "; key$key=...\n" if $key > 25;
+		push @rdata, unpack 'H4H4',    pack( 'n2', $key, length $val );
 		push @rdata, split /(\S{32})/, unpack 'H*', $val;
 		$length += 4 + length $val;
 	}
+	if ( $self->{unparsed} ) {
+		my @hex = split /(\S{32})/, unpack 'H*', $self->{unparsed};
+		push @rdata, "\n", shift(@hex), "\t; corrupt RDATA\n", @hex;
+	}
 	return ( "\\# $length", @rdata );
 }
 
 
 sub _parse_rdata {			## populate RR from rdata in argument list
-	my $self = shift;
+	my ( $self, @argument ) = @_;
 
-	$self->svcpriority(shift);
-	$self->targetname(shift);
+	$self->svcpriority( shift @argument );
+	$self->targetname( shift @argument );
 
 	local $SIG{__WARN__} = sub { die @_ };
-	while ( my $svcparam = shift ) {
+	while ( my $svcparam = shift @argument ) {
 		for ($svcparam) {
 			my @value;
 			if (/^key\d+=(.*)$/i) {
-				push @value, length($1) ? $1 : shift;
-			} elsif (/=(.*)$/) {
-				local $_ = length($1) ? $1 : shift;
-				s/^"(.*)"$/$1/;			# strip enclosing quotes
+				push @value, length($1) ? $1 : shift @argument;
+			} elsif (/^[^=]+=(.*)$/) {
+				local $_ = length($1) ? $1 : shift @argument;
+				s/^"([^"]*)"$/$1/;		# strip enclosing quotes
 				push @value, split /,/;
 			} else {
 				push @value, '' unless $keybyname{lc $_};    # empty | Boolean
@@ -131,19 +140,20 @@ sub _parse_rdata {			## populate RR from rdata in argument list
 sub _post_parse {			## parser post processing
 	my $self = shift;
 
-	my $params = $self->{SvcParams} || return;
-	my %params = @$params;
+	my $paramref = $self->{SvcParams} || [];
+	my %svcparam = scalar(@$paramref) ? @$paramref : return;
+
 	$self->key0(undef);					# ruse to force sorting of SvcParams
-	if ( defined $params{0} ) {
+	if ( defined $svcparam{0} ) {
 		my %unique;
-		foreach ( grep { !$unique{$_}++ } unpack 'n*', $params{0} ) {
-			croak( $self->type . qq[: unexpected "key0" in mandatory list] ) if $unique{0};
-			croak( $self->type . qq[: duplicate "key$_" in mandatory list] ) if --$unique{$_};
-			croak( $self->type . qq[: mandatory "key$_" not present] ) unless defined $params{$_};
+		foreach ( grep { !$unique{$_}++ } unpack 'n*', $svcparam{0} ) {
+			die( $self->type . qq[: unexpected "key0" in mandatory list] ) if $unique{0};
+			die( $self->type . qq[: duplicate "key$_" in mandatory list] ) if --$unique{$_};
+			die( $self->type . qq[: mandatory "key$_" not present] ) unless defined $svcparam{$_};
 		}
 		$self->mandatory( keys %unique );		# restore mandatory key list
 	}
-	croak( $self->type . qq[: expected alpn="..." not present] ) if defined( $params{2} ) and !$params{1};
+	die( $self->type . qq[: expected alpn="..." not present] ) if defined( $svcparam{2} ) && !$svcparam{1};
 	return;
 }
 
@@ -157,17 +167,16 @@ sub _defaults {				## specify RR attribute default values
 
 
 sub svcpriority {
-	my $self = shift;					# uncoverable pod
-
-	$self->{SvcPriority} = 0 + shift if scalar @_;
+	my ( $self, @value ) = @_;				# uncoverable pod
+	for (@value) { $self->{SvcPriority} = 0 + $_ }
 	return $self->{SvcPriority} || 0;
 }
 
 
 sub targetname {
-	my $self = shift;					# uncoverable pod
+	my ( $self, @value ) = @_;				# uncoverable pod
 
-	$self->{TargetName} = Net::DNS::DomainName->new(shift) if scalar @_;
+	for (@value) { $self->{TargetName} = Net::DNS::DomainName->new($_) }
 
 	my $target = $self->{TargetName} ? $self->{TargetName}->name : return;
 	return $target unless $self->{SvcPriority};
@@ -176,40 +185,50 @@ sub targetname {
 
 
 sub mandatory {				## mandatory=key1,port,...
-	my $self = shift;
-	my @list = map { $keybyname{lc $_} || $_ } map { split /,/ } @_;
-	my @keys = map { /(\d+)$/ ? $1 : croak( $self->type . qq[: unexpected "$_"] ) } @list;
+	my ( $self, @value ) = @_;
+	my @list = map { $keybyname{lc $_} || $_ } map { split /,/ } @value;
+	my @keys = map { /(\d+)$/ ? $1 : die( $self->type . qq[: unexpected "$_"] ) } @list;
 	return $self->key0( _integer16( sort { $a <=> $b } @keys ) );
 }
 
 sub alpn {				## alpn=h3,h2,...
-	my $self = shift;
-	return $self->key1( _string(@_) );
+	my ( $self, @value ) = @_;
+	return $self->key1( _string(@value) );
 }
 
-sub no_default_alpn {			## no-default-alpn
-	my $self = shift;					# uncoverable pod
-	return $self->key2( ( defined(wantarray) ? @_ : '' ), @_ );
+sub no_default_alpn {			## no-default-alpn	(Boolean)
+	my ( $self, @value ) = @_;				# uncoverable pod
+	return $self->key2( ( defined(wantarray) ? () : '' ), @value );
 }
 
 sub port {				## port=1234
-	my $self = shift;
-	return $self->key3( map { _integer16($_) } @_ );
+	my ( $self, @value ) = @_;
+	return $self->key3( map { _integer16($_) } @value );
 }
 
 sub ipv4hint {				## ipv4hint=192.0.2.1,...
-	my $self = shift;
-	return $self->key4( _ipv4(@_) );
+	my ( $self, @value ) = @_;
+	return $self->key4( _ipv4(@value) );
 }
 
-sub ech {				## ech=base64string
-	my $self = shift;
-	return $self->key5( map { _base64($_) } @_ );
+sub ech {				## Format not specified
+	my ( $self, @value ) = @_;
+	return $self->key5(@value);				# RESERVED
 }
 
 sub ipv6hint {				## ipv6hint=2001:DB8::1,...
-	my $self = shift;
-	return $self->key6( _ipv6(@_) );
+	my ( $self, @value ) = @_;
+	return $self->key6( _ipv6(@value) );
+}
+
+sub dohpath {				## dohpath=/dns-query{?dns}
+	my ( $self, @value ) = @_;				# uncoverable pod
+	return $self->key7(@value);
+}
+
+sub ohttp {				## ohttp	(Boolean)
+	my ( $self, @value ) = @_;				# uncoverable pod
+	return $self->key8( ( defined(wantarray) ? () : '' ), @value );
 }
 
 
@@ -217,66 +236,69 @@ sub ipv6hint {				## ipv6hint=2001:DB8::1,...
 
 
 sub _presentation {			## render octet string(s) in presentation format
-	return () unless scalar @_;
-	my $raw = join '', @_;
+	my @arg = @_;
+	my $raw = scalar(@arg) ? join( '', @arg ) : return ();
 	return Net::DNS::Text->decode( \$raw, 0, length($raw) )->string;
 }
 
-sub _base64 {
-	return _presentation( map { MIME::Base64::decode($_) } @_ );
-}
-
 sub _integer16 {
-	return _presentation( map { pack( 'n', $_ ) } @_ );
+	my @arg = @_;
+	return _presentation( map { pack( 'n', $_ ) } @arg );
 }
 
 sub _ipv4 {
-	return _presentation( map { Net::DNS::RR::A::address( {}, $_ ) } @_ );
+	my @arg = @_;
+	return _presentation( map { Net::DNS::RR::A::address( {}, $_ ) } @arg );
 }
 
 sub _ipv6 {
-	return _presentation( map { Net::DNS::RR::AAAA::address( {}, $_ ) } @_ );
+	my @arg = @_;
+	return _presentation( map { Net::DNS::RR::AAAA::address( {}, $_ ) } @arg );
 }
 
 sub _string {
-	local $_ = join ',', '', @_;				# reassemble argument string
-	s/\\092,/\\044/g;		### tolerate unnecessary double-escape nonsense in
-	s/\\092\\092/\\092/g;		### draft-ietf-dnsop-svcb-https that contradicts RFC1035
+	my @arg = @_;
+	local $_ = join ',', @arg;				# reassemble argument string
 	s/\\,/\\044/g;						# disguise (RFC1035) escaped comma
-	my ( undef, @reparsed ) = split /,/;			# multi-valued argument
-	return _presentation( map { Net::DNS::Text->new($_)->encode() } @reparsed );
+	die <<"QQ" if /\\092,|\\092\\092/;
+SVCB:	Please use standard RFC1035 escapes
+	RFC9460 double-escape nonsense not implemented
+QQ
+	return _presentation( map { Net::DNS::Text->new($_)->encode() } split /,/ );
 }
 
 
-our $AUTOLOAD;
-
 sub AUTOLOAD {				## Dynamic constructor/accessor methods
-	my $self = shift;
+	my ( $self, @argument ) = @_;
 
+	our $AUTOLOAD;
 	my ($method) = reverse split /::/, $AUTOLOAD;
 
-	my $default = join '::', 'SUPER', $method;
-	return $self->$default(@_) unless $method =~ /^key[0]*(\d+)$/i;
+	my $super = "SUPER::$method";
+	return $self->$super(@argument) unless $method =~ /^key[0]*(\d+)$/i;
 	my $key = $1;
 
-	my $params = $self->{SvcParams} || [];
-	my %params = @$params;
+	my $paramsref = $self->{SvcParams} || [];
+	my %svcparams = @$paramsref;
 
-	if ( scalar @_ ) {
-		my $arg = shift;				# keyNN($value);
-		delete $params{$key} unless defined $arg;
-		croak( $self->type . qq[: duplicate SvcParam "key$key"] ) if defined $params{$key};
-		$params{$key} = Net::DNS::Text->new("$arg")->raw if defined $arg;
-		$self->{SvcParams} = [map { ( $_, $params{$_} ) } sort { $a <=> $b } keys %params];
-		croak( $self->type . qq[: unexpected number of values for "key$key"] ) if scalar @_;
+	if ( scalar @argument ) {
+		my $arg = shift @argument;			# keyNN($value);
+		delete $svcparams{$key} unless defined $arg;
+		die( $self->type . qq[: duplicate SvcParam "key$key"] ) if defined $svcparams{$key};
+		die( $self->type . qq[: invalid SvcParam "key$key"] )	if $key > 65534;
+		$svcparams{$key}   = Net::DNS::Text->new("$arg")->raw if defined $arg;
+		$self->{SvcParams} = [map { ( $_, $svcparams{$_} ) } sort { $a <=> $b } keys %svcparams];
+		die( $self->type . qq[: unexpected number of values for "key$key"] ) if scalar @argument;
 	} else {
-		croak( $self->type . qq[: no value specified for "key$key"] ) unless defined wantarray;
+		die( $self->type . qq[: no value specified for "key$key"] ) unless defined wantarray;
 	}
 
-	my $value = $params{$key};
+	my $value = $svcparams{$key};
 	return defined($value) ? _presentation($value) : $value;
 }
 
+########################################
+
 
 1;
 __END__
@@ -285,7 +307,7 @@ __END__
 =head1 SYNOPSIS
 
     use Net::DNS;
-    $rr = Net::DNS::RR->new('name HTTPS SvcPriority TargetName alpn=h3,...');
+    $rr = Net::DNS::RR->new('name HTTPS SvcPriority TargetName SvcParams');
 
 =head1 DESCRIPTION
 
@@ -329,13 +351,13 @@ owner name of this record must be used as the effective TargetName.
 
 =head2 mandatory, alpn, no-default-alpn, port, ipv4hint, ech, ipv6hint
 
-    $rr = Net::DNS::RR->new( 'svc.example. SVCB 1 svc.example. port=1234' );
+    $rr = Net::DNS::RR->new( 'svcb.example. SVCB 1 svcb.example. port=1234' );
 
     $rr->port(1234);
     $string = $rr->port();	# \004\210
     $rr->key3($string);
 
-Constructor methods for mnemonic SvcParams defined in draft-ietf-dnsop-svcb-https.
+Constructor methods for mnemonic SvcParams prescribed by RFC9460.
 When invoked without arguments, the methods return the presentation format
 value for the underlying key.
 The behaviour with undefined arguments is not specified.
@@ -355,7 +377,7 @@ The specified key will be deleted if the value is undefined.
 
 =head1 COPYRIGHT
 
-Copyright (c)2020-2021 Dick Franks. 
+Copyright (c)2020-2022 Dick Franks. 
 
 All rights reserved.
 
@@ -366,7 +388,7 @@ Package template (c)2009,2012 O.M.Kolkman and R.W.Franks.
 
 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose and without fee is hereby granted, provided
-that the above copyright notice appear in all copies and that both that
+that the original copyright notices appear in all copies and that both
 copyright notice and this permission notice appear in supporting
 documentation, and that the name of the author not be used in advertising
 or publicity pertaining to distribution of the software without specific
@@ -383,6 +405,9 @@ DEALINGS IN THE SOFTWARE.
 
 =head1 SEE ALSO
 
-L<perl>, L<Net::DNS>, L<Net::DNS::RR>, draft-ietf-dnsop-svcb-https
+L<perl> L<Net::DNS> L<Net::DNS::RR>
+L<RFC9460|https://tools.ietf.org/html/rfc9460>
+
+L<Service Parameter Keys|https://www.iana.org/assignments/dns-svcb>
 
 =cut
